use std::net::SocketAddr;

use tonic::transport::Server;

use crate::{
    app::SharedApp,
    services::release_manager::{
        extensions::ReleaseManagerExt, models::CommitArtifact, ReleaseManager,
    },
};

use self::gen::{greeter_server, HelloReply, HelloRequest};

mod gen {
    tonic::include_proto!("flux_releaser");
}

pub struct FluxReleaserGrpc {
    release_manager: ReleaseManager,
}

impl FluxReleaserGrpc {
    pub fn new(app: SharedApp) -> Self {
        Self {
            release_manager: app.release_manager(),
        }
    }
}

#[tonic::async_trait]
impl greeter_server::Greeter for FluxReleaserGrpc {
    async fn say_hello(
        &self,
        request: tonic::Request<HelloRequest>,
    ) -> std::result::Result<tonic::Response<HelloReply>, tonic::Status> {
        let req = request.into_inner();
        self.release_manager
            .commit_artifact(
                req.try_into()
                    .map_err(|e: anyhow::Error| tonic::Status::invalid_argument(e.to_string()))?,
            )
            .await
            .unwrap();

        Ok(tonic::Response::new(HelloReply {
            message: "something".into(),
        }))
    }
}

impl TryFrom<HelloRequest> for CommitArtifact {
    type Error = anyhow::Error;

    fn try_from(value: HelloRequest) -> Result<Self, Self::Error> {
        if value.app.is_empty() {
            anyhow::bail!("app cannot be empty")
        }
        if value.branch.is_empty() {
            anyhow::bail!("branch cannot be empty")
        }
        if value.folder.is_empty() {
            anyhow::bail!("folder cannot be empty")
        }

        Ok(Self {
            app: value.app,
            branch: value.branch,
            folder: value.folder.into(),
        })
    }
}

pub async fn tonic_serve(host: SocketAddr, app: SharedApp) -> anyhow::Result<()> {
    tracing::info!("grpc listening on: {}", host);
    Server::builder()
        .add_service(greeter_server::GreeterServer::new(FluxReleaserGrpc::new(
            app,
        )))
        .serve(host)
        .await?;

    Ok(())
}
